from acherion import (
    AcherionGraph,
    AcherionNode,
    acherion_plotly_template_payload,
    acherion_ui_colors,
    build_acherion_theme_override_css,
    build_embedded_acherion_theme_css,
    compile_acherion_graph,
    custom_function_params,
    execute_acherion_graph,
    normalize_acherion_theme_overrides,
)
from acherion.assets.js.graph import GRAPH_JS
from acherion.assets.js.viewport import VIEWPORT_JS
from acherion.catalog import modules as _catalog_modules
from acherion.catalog import runtime as _catalog_runtime
from acherion.embed.designer.interactions import _DesignerInteractionsMixin
import acherion.embed.render.editor as acherion_render_editor
from acherion.embed.render.pins import _RenderPinsMixin
from acherion.graphops.catalog import _GraphOpsCatalogMixin
from acherion.graphops.pins import _GraphOpsPinsMixin
from acherion.preview import preview_value_plotly_payload
import numpy as np
import plotly.graph_objects as go
from acherion.standalone_host import (
    load_acherion_graph_namespace,
    run_standalone_acherion_preview,
    standalone_editor_visible_code,
)


def test_constant_graph_compiles_executes_and_previews() -> None:
    graph = AcherionGraph(
        nodes=[
            AcherionNode(
                node_id='n1',
                kind='constant',
                params={
                    'value_type': 'int',
                    'number_value': 3,
                },
            )
        ]
    )

    source_code = compile_acherion_graph(graph)

    assert 'def run(bindings=None):' in source_code

    local_values = execute_acherion_graph(source_code)

    assert 3 in local_values.values()

    preview = run_standalone_acherion_preview(graph)

    assert preview.reference_values['n1'] == 3
    assert preview.reference_values['n1@0'] == 3
    assert preview.state_values['acherion.preview_value_count'] >= 1
    assert str(preview.state_values.get('hint') or '').startswith(
        'Preview ran for '
    )


def test_duplicate_titled_producer_nodes_compile_to_distinct_variables() -> None:
    graph = AcherionGraph(
        nodes=[
            AcherionNode(
                node_id='c1',
                kind='constant',
                title='One',
                params={
                    'value_type': 'int',
                    'number_value': 1,
                },
            ),
            AcherionNode(
                node_id='c2',
                kind='constant',
                title='Two',
                params={
                    'value_type': 'int',
                    'number_value': 2,
                },
            ),
            AcherionNode(
                node_id='c3',
                kind='constant',
                title='Ten',
                params={
                    'value_type': 'int',
                    'number_value': 10,
                },
            ),
            AcherionNode(
                node_id='c4',
                kind='constant',
                title='Twenty',
                params={
                    'value_type': 'int',
                    'number_value': 20,
                },
            ),
            AcherionNode(
                node_id='a1',
                kind='op_arithmetic',
                title='Arithmetic',
                params={
                    'operator': '+',
                    'left_source': 'c1',
                    'right_source': 'c2',
                },
            ),
            AcherionNode(
                node_id='a2',
                kind='op_arithmetic',
                title='Arithmetic',
                params={
                    'operator': '+',
                    'left_source': 'c3',
                    'right_source': 'c4',
                },
            ),
            AcherionNode(
                node_id='m1',
                kind='make_list',
                title='Pair',
                params={
                    'arg_sources': ['a1', 'a2'],
                },
            ),
        ]
    )

    source_code = compile_acherion_graph(graph)
    local_values = execute_acherion_graph(source_code)

    assert 'arithmetic_5 = (one_1) + (two_2)' in source_code
    assert 'arithmetic_6 = (ten_3) + (twenty_4)' in source_code
    assert local_values['pair_7'] == [3, 30]



def test_standalone_visible_code_hides_runtime_scaffold() -> None:
    graph = AcherionGraph(
        nodes=[
            AcherionNode(
                node_id='n1',
                kind='constant',
                params={
                    'value_type': 'bool',
                    'bool_value': True,
                },
            )
        ]
    )

    source_code = compile_acherion_graph(graph)
    visible_code = standalone_editor_visible_code(source_code)

    assert '_normalize_scoped_bindings' not in visible_code
    assert 'ACHERION_EXTERNAL_EVENTS' not in visible_code
    assert 'def run():' in visible_code
    assert 'bindings=None' not in visible_code


def test_theme_helpers_normalize_aliases_and_preserve_scope_contract() -> None:
    overrides = normalize_acherion_theme_overrides({
        'primary': '#123456',
        'sidebar_panel_bg': 'rgba(1, 2, 3, 0.5)',
        '--custom-gap': '10px',
        'muted': '',
    })

    assert overrides == {
        '--oe-blue': '#123456',
        '--ach-sidebar-panel-bg': 'rgba(1, 2, 3, 0.5)',
        '--custom-gap': '10px',
    }

    colors = acherion_ui_colors({'primary': '#123456'})

    assert colors['primary'] == '#123456'
    assert colors['positive']

    override_css = build_acherion_theme_override_css(
        scope_selector='.acherion-scope',
        theme_overrides={'primary': '#123456'},
    )
    embedded_css = build_embedded_acherion_theme_css(
        scope_selector='.acherion-scope',
        theme_overrides={'primary': '#123456'},
    )

    assert '.acherion-scope {' in override_css
    assert '--oe-blue: #123456;' in override_css
    assert '.acherion-scope' in embedded_css
    assert '--oe-blue: #123456;' in embedded_css


def test_function_catalog_includes_numpy_and_plotly_modules() -> None:
    module_options = _catalog_modules.module_options()

    assert module_options['np'] == 'numpy (np)'
    assert module_options['go'] == 'plotly.graph_objects (go)'
    assert module_options['px'] == 'plotly.express (px)'
    assert module_options['plotly.subplots'] == 'plotly.subplots'

    np_options = _catalog_runtime.func_options('np')
    go_options = _catalog_runtime.func_options('go')
    px_options = _catalog_runtime.func_options('px')
    subplot_options = _catalog_runtime.func_options('plotly.subplots')

    assert 'np.array' in np_options
    assert 'go.Figure' in go_options
    assert 'px.scatter' in px_options
    assert 'plotly.subplots.make_subplots' in subplot_options


def test_standalone_runtime_exposes_catalog_module_globals() -> None:
    source_code = '''
def run(bindings=None):
    array = np.array([1, 2, 3])
    root = math.sqrt(9)
    figure = go.Figure()
    subplot = plotly.subplots.make_subplots(rows=1, cols=1)
    return {
        "array": array,
        "root": root,
        "figure": figure,
        "subplot": subplot,
    }
'''

    local_values = execute_acherion_graph(source_code)

    assert local_values['array'].tolist() == [1, 2, 3]
    assert local_values['root'] == 3.0
    assert type(local_values['figure']).__name__ == 'Figure'
    assert type(local_values['subplot']).__name__ == 'Figure'


def test_clear_selection_shortcut_clears_selected_nodes() -> None:
    class _StubDesigner(_DesignerInteractionsMixin):
        def __init__(self) -> None:
            self._selected_node_ids = {'n1', 'n2'}
            self._selected_connection_id = None
            self.refresh_count = 0
            self.hint_updates = 0

        def refresh(self) -> None:
            self.refresh_count += 1

        def _update_hint(self) -> None:
            self.hint_updates += 1

    designer = _StubDesigner()

    class _Event:
        args = {'shortcut_id': 'clear_selection', 'key': 'Escape'}

    designer._handle_canvas_key(_Event())

    assert designer._selected_node_ids == set()
    assert designer.refresh_count == 1
    assert designer.hint_updates == 1


def _capture_editor_number_call(monkeypatch, node: AcherionNode) -> dict[str, object]:
    calls: list[dict[str, object]] = []

    class _StubUiElement:
        def props(self, _value):
            return self

        def classes(self, _value):
            return self

    def _fake_number(*args, **kwargs):
        calls.append({'args': args, 'kwargs': kwargs})
        return _StubUiElement()

    class _StubEditor(acherion_render_editor._RenderEditorMixin):
        def __init__(self) -> None:
            self._host = None

        def _is_system_sink_node(self, _node: AcherionNode) -> bool:
            return False

    monkeypatch.setattr(
        acherion_render_editor,
        'get_acherion_node_definition',
        lambda _kind: None,
    )
    monkeypatch.setattr(
        acherion_render_editor.ui,
        'input',
        lambda *args, **kwargs: _StubUiElement(),
    )
    monkeypatch.setattr(
        acherion_render_editor.ui,
        'label',
        lambda *args, **kwargs: _StubUiElement(),
    )
    monkeypatch.setattr(acherion_render_editor.ui, 'number', _fake_number)

    _StubEditor()._render_node_config_fields(node)

    assert calls
    return calls[0]


def test_make_list_editor_uses_numeric_number_bounds(monkeypatch) -> None:
    call = _capture_editor_number_call(
        monkeypatch,
        AcherionNode(
            node_id='list',
            kind='make_list',
            params={'arg_count': 3},
        ),
    )

    kwargs = call['kwargs']

    assert kwargs['min'] == 0
    assert isinstance(kwargs['min'], int)
    assert kwargs['step'] == 1
    assert isinstance(kwargs['step'], int)


def test_sequencer_editor_uses_numeric_number_bounds(monkeypatch) -> None:
    call = _capture_editor_number_call(
        monkeypatch,
        AcherionNode(
            node_id='seq',
            kind='sequencer',
            params={'then_count': 4},
        ),
    )

    kwargs = call['kwargs']

    assert kwargs['min'] == 2
    assert isinstance(kwargs['min'], int)
    assert kwargs['max'] == 8
    assert isinstance(kwargs['max'], int)
    assert kwargs['step'] == 1
    assert isinstance(kwargs['step'], int)


def test_client_js_batches_connection_updates_in_hot_paths() -> None:
    assert (
        "return JSON.stringify(String(value ?? '')).slice(1, -1);"
        in GRAPH_JS
    )
    assert "String(pinIndex ?? '')" in GRAPH_JS
    assert 'queueConnectionUpdate(stage)' in GRAPH_JS
    assert 'stage.__oeAcherionConnectionUpdateQueued' in GRAPH_JS
    assert 'connectionPaths(stage, nodeIds)' in GRAPH_JS
    assert 'if (seen.has(path)) return;' in GRAPH_JS
    assert 'path.className ||' not in GRAPH_JS
    assert 'this.queueConnectionUpdate(stage, [nodeId]);' in GRAPH_JS
    assert 'const pinCenter = (nodeId, direction, pinIndex) => {' in GRAPH_JS
    assert 'this.queueConnectionUpdate(stage);' not in VIEWPORT_JS


def test_plot_figure_codegen_uses_shared_go_runtime_global() -> None:
    graph = AcherionGraph(
        nodes=[
            AcherionNode(
                node_id='x1',
                kind='constant',
                params={
                    'value_type': 'int',
                    'number_value': 1,
                },
            ),
            AcherionNode(
                node_id='x2',
                kind='constant',
                params={
                    'value_type': 'int',
                    'number_value': 2,
                },
            ),
            AcherionNode(
                node_id='xs',
                kind='make_list',
                title='X Values',
                params={
                    'arg_sources': ['x1', 'x2'],
                },
            ),
            AcherionNode(
                node_id='y1',
                kind='constant',
                params={
                    'value_type': 'int',
                    'number_value': 3,
                },
            ),
            AcherionNode(
                node_id='y2',
                kind='constant',
                params={
                    'value_type': 'int',
                    'number_value': 4,
                },
            ),
            AcherionNode(
                node_id='ys',
                kind='make_list',
                title='Y Values',
                params={
                    'arg_sources': ['y1', 'y2'],
                },
            ),
            AcherionNode(
                node_id='fig',
                kind='plot_figure',
                title='Figure',
                params={
                    'figure_type': 'scatter',
                    'named_sources': {
                        'x': 'xs',
                        'y': 'ys',
                    },
                    'figure_title': 'Demo',
                },
            ),
        ]
    )

    source_code = compile_acherion_graph(graph)

    assert 'import plotly.graph_objects as _go' not in source_code
    assert 'go.Figure(' in source_code
    assert 'go.Scatter(' in source_code

    preview = run_standalone_acherion_preview(graph)
    figure = preview.reference_values['fig']

    assert isinstance(figure, go.Figure)
    assert tuple(figure.data[0].x) == (1, 2)
    assert tuple(figure.data[0].y) == (3, 4)
    assert figure.layout.title.text == 'Demo'


def test_plotly_template_payload_matches_dark_acherion_tokens() -> None:
    template = acherion_plotly_template_payload()

    assert template['layout']['paper_bgcolor'] == 'rgba(0,0,0,0)'
    assert template['layout']['font']['color'] == '#E7E9EA'
    assert template['layout']['legend']['bgcolor'] == 'rgba(22, 24, 28, 0.85)'
    assert template['layout']['colorway'][:4] == [
        '#1D9BF0',
        '#00BA7C',
        '#F4212E',
        '#FFD400',
    ]


def test_plotly_preview_payload_serializes_figure_like_values() -> None:
    figure = go.Figure(
        data=[go.Bar(x=[1, 2], y=[3, 4])],
        layout={'title': {'text': 'Preview'}},
    )

    payload = preview_value_plotly_payload(figure)

    assert payload is not None
    assert payload['data'][0]['type'] == 'bar'
    assert payload['layout']['title']['text'] == 'Preview'
    assert payload['layout']['template']['layout']['font']['color'] == '#E7E9EA'


def test_numpy_dependency_is_available_for_array_summaries() -> None:
    value = np.array([[1, 2], [3, 4]])

    assert value.shape == (2, 2)


def test_int_typed_pin_literals_are_coerced_before_preview_runtime() -> None:
    class _DummyPinRenderer(_RenderPinsMixin):
        def __init__(self) -> None:
            self.change_count = 0

        def _notify_change(self) -> None:
            self.change_count += 1

    renderer = _DummyPinRenderer()
    node = AcherionNode(
        node_id='call',
        kind='call_function',
        params={'pin_literals': {}},
    )

    renderer._set_pin_literal_value(
        node,
        pin_id='arg:2',
        input_kind='number',
        pin_type='int',
        value=250.0,
    )
    renderer._set_pin_literal_value(
        node,
        pin_id='arg:3',
        input_kind='number',
        pin_type='float',
        value=2,
    )

    assert node.params['pin_literals']['arg:2'] == 250
    assert isinstance(node.params['pin_literals']['arg:2'], int)
    assert node.params['pin_literals']['arg:3'] == 2.0
    assert isinstance(node.params['pin_literals']['arg:3'], float)
    assert renderer.change_count == 2


def test_custom_function_parser_accepts_plain_function_signature() -> None:
    class _DummyCatalog(_GraphOpsCatalogMixin):
        pass

    data, error = _DummyCatalog()._parse_custom_function_source(
        'def custom_plain(x: int = 2):\n'
        '    return x + 1\n'
    )

    assert error == ''
    assert data is not None
    assert data['signature'] == 'custom_plain(x: int = 2)'
    assert data['min_args'] == 0
    assert data['max_args'] == 1
    assert data['param_names'] == ['x']
    assert data['param_types'] == ['int']


def test_custom_function_unknown_non_none_return_still_exposes_output_pin() -> None:
    class _DummyCustomFunctionOwner(_GraphOpsCatalogMixin, _GraphOpsPinsMixin):
        def __init__(self) -> None:
            self._graph = AcherionGraph(nodes=[], user_functions={})

    owner = _DummyCustomFunctionOwner()
    function_path = 'user.custom_function_1'
    data, error = owner._parse_custom_function_source(
        'def custom_function_1(x):\n'
        '    return x + 1\n'
    )

    assert error == ''
    assert data is not None
    assert data['return_type'] == 'any'

    owner._graph.user_functions[function_path] = data
    pins = owner._callable_output_pin_specs(
        AcherionNode(
            node_id='cf1',
            kind='custom_function',
            params={'function_path': function_path},
        )
    )

    assert pins == [{'pin_id': 'value', 'label': 'result', 'type': 'any'}]


def test_new_custom_functions_get_sequential_default_names() -> None:
    class _DummyCustomFunctionOwner(_GraphOpsCatalogMixin):
        def __init__(self) -> None:
            self._graph = AcherionGraph(nodes=[], user_functions={})

    owner = _DummyCustomFunctionOwner()
    first = AcherionNode(
        node_id='cf1',
        kind='custom_function',
        title='Custom Function',
        params=custom_function_params('cf1'),
    )
    owner._ensure_custom_function_entry(first)
    owner._graph.nodes.append(first)

    second = AcherionNode(
        node_id='cf2',
        kind='custom_function',
        title='Custom Function',
        params=custom_function_params('cf2'),
    )
    owner._ensure_custom_function_entry(second)

    assert first.params['function_path'] == 'user.custom_function_1'
    assert second.params['function_path'] == 'user.custom_function_2'
    assert 'user.custom_function_1' in owner._graph.user_functions
    assert 'user.custom_function_2' in owner._graph.user_functions


def test_legacy_self_custom_function_runtime_runs_without_migration() -> None:
    function_path = 'user.custom_function_a3ca1e8c'
    graph = AcherionGraph(
        nodes=[],
        user_functions={
            function_path: {
                'label': 'custom_function_a3ca1e8c',
                'signature': 'custom_function_a3ca1e8c(x)',
                'min_args': 1,
                'max_args': 1,
                'param_names': ['x'],
                'param_types': ['int'],
                'return_type': 'int',
                'source_code': (
                    'def custom_function_a3ca1e8c(self, x):\n'
                    '    return x + 1\n'
                ),
            }
        },
    )

    source_code = compile_acherion_graph(graph)
    namespace = load_acherion_graph_namespace(source_code)

    assert "return __acherion_fn(globals().get('self'), *args)" in source_code
    assert namespace['custom_function_a3ca1e8c'](2) == 3


"""Shell and lifecycle mixin for AcherionDesigner."""

# pyright: reportGeneralTypeIssues=false

from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Any, Callable, Literal, cast

from nicegui import background_tasks, ui

from acherion.assets import (
    _ACH_CLIENT_JS,
    _ACH_CSS,
)
from acherion.events import (
    collect_acherion_external_events,
    serializable_acherion_external_events,
)
from acherion.model import (
    AcherionGraph,
    AcherionNode,
    _graph_from_dict,
    _graph_to_dict,
)
from acherion.preview import (
    AcherionPreviewRunResult,
)

if TYPE_CHECKING:
    from nicegui.client import Client


class _DesignerShellMixin:
    """Public API, build, refresh, and notification helpers."""

    _graph: AcherionGraph
    _drag_node_id: str | None = None
    _pending_source_node_id: str | None = None
    _status_override_text: str | None = None
    _status_override_negative: bool = False
    _selected_connection_id: Any = None

    @property
    def mode(self: Any) -> str:
        """Return current Logic Studio view mode."""
        return str(self._editor_mode)

    @property
    def nodes(self) -> list[AcherionNode]:
        """Return the current graph nodes."""
        return self._graph.nodes

    def has_nodes(self: Any) -> bool:
        """Return whether the graph contains any nodes."""
        return bool(self._graph.nodes)

    def has_user_logic(self: Any) -> bool:
        """Return whether graph contains user-authored logic."""
        if any(not self._is_system_node(n) for n in self._graph.nodes):
            return True
        return any(
            self._is_system_sink_node(n)
            and str(n.params.get('source') or '').strip()
            for n in self._graph.nodes
        )

    def _normalize_graph(self: Any) -> None:
        self._sync_manual_schema_keys()
        self._prune_obsolete_function_io_nodes()
        self._sync_system_nodes()
        self._ensure_function_box_entries()
        self._sync_custom_function_nodes()
        self._prune_invalid_exec_connections()

    def set_graph_from_dict(self: Any, data: dict[str, Any] | None) -> None:
        """Replace graph state from persisted builder_state data."""
        self._graph = _graph_from_dict(data)
        self._normalize_graph()
        self.clear_preview_results()
        self.force_redraw()

    def to_dict(self: Any) -> dict[str, Any]:
        """Return the current graph as a serialisable dict."""
        self._normalize_graph()
        return _graph_to_dict(self._graph)

    def graph_state(self: Any) -> AcherionGraph:
        """Return a detached normalized graph snapshot."""
        return _graph_from_dict(self.to_dict())

    def preview_bindings(self: Any) -> dict[str, dict[str, Any]]:
        """Return current transient preview values grouped by scope."""
        return {
            scope: dict(values)
            for scope, values in self._preview_bindings.items()
        }

    def preview_binding_values(self: Any, scope: str) -> dict[str, Any]:
        """Return current transient preview values for one scope."""
        clean_scope = str(scope or '').strip()
        if not clean_scope:
            return {}
        return dict(self._preview_bindings.get(clean_scope, {}))

    def preview_reference_values(self: Any) -> dict[str, Any]:
        """Return latest preview values keyed by opaque runtime reference."""
        return dict(self._preview_reference_values)

    def preview_state_values(self: Any) -> dict[str, Any]:
        """Return latest preview state keyed by host-defined names."""
        return dict(self._preview_state_values)

    def set_preview_binding_value(
        self: Any,
        scope: str,
        key: str,
        value: Any,
    ) -> None:
        """Store one transient preview value by scope and schema key."""
        clean_scope = str(scope or '').strip()
        clean_key = str(key or '').strip()
        if not clean_scope or not clean_key:
            return
        scope_values = self._preview_bindings.setdefault(clean_scope, {})
        scope_values[clean_key] = value
        self._clear_preview_runtime_state()
        self.refresh()
        if self._on_run_preview is not None:
            self._on_run_preview()

    def clear_preview_binding_value(self: Any, scope: str, key: str) -> None:
        """Remove one transient preview value by scope and key."""
        clean_scope = str(scope or '').strip()
        clean_key = str(key or '').strip()
        if not clean_scope or not clean_key:
            return
        scope_values = self._preview_bindings.get(clean_scope)
        if scope_values is not None:
            scope_values.pop(clean_key, None)
            if not scope_values:
                self._preview_bindings.pop(clean_scope, None)
        self._clear_preview_runtime_state()
        self.refresh()
        if self._on_run_preview is not None:
            self._on_run_preview()

    def apply_preview_result(
        self: Any,
        result: AcherionPreviewRunResult,
    ) -> None:
        """Replace the latest runtime preview result for the current graph."""
        self._preview_reference_values = dict(result.reference_values)
        self._preview_state_values = dict(result.state_values)
        self.refresh()
        hint = str(self._preview_state_values.get('hint') or '').strip()
        self._update_hint(hint or 'Preview updated.')

    def clear_preview_results(self: Any) -> None:
        """Clear all transient preview bindings and runtime preview state."""
        self._preview_bindings = {}
        self._clear_preview_runtime_state()

    def _clear_preview_runtime_state(self: Any) -> None:
        """Clear preview values captured from the last preview execution."""
        self._preview_reference_values = {}
        self._preview_state_values = {}

    def generated_user_code(self: Any) -> str:
        """Return the current graph compiled as editor-ready user code."""
        self._normalize_graph()
        if self._host is not None:
            return str(self._host.generated_user_code(self))
        raise RuntimeError(
            'AcherionDesigner requires a host to generate user code.'
        )

    def generated_runtime_bindings(
        self: Any,
    ) -> dict[str, dict[str, dict[str, Any]]]:
        """Return hidden UI-role bindings for generated graph code."""
        self._normalize_graph()
        if self._host is not None:
            return cast(
                dict[str, dict[str, dict[str, Any]]],
                self._host.generated_runtime_bindings(self),
            )
        return {}

    def generated_external_events(
        self: Any,
    ) -> dict[str, dict[str, str]]:
        """Return hidden external-event metadata for generated graph code."""
        self._normalize_graph()
        return serializable_acherion_external_events(
            collect_acherion_external_events(self._graph.nodes)
        )

    def build(self: Any) -> None:
        """Render the visual-logic designer into current NiceGUI context."""
        self._client = ui.context.client
        self._ensure_css()
        self._ensure_client_js()
        self._normalize_graph()
        with ui.element('div').classes(
            'ach-workbench oe-surface rounded-lg w-full min-w-0'
        ).props(f'id={self._frame_dom_id}'):
            self._render_workbench_menu_bar()
            with ui.element('div').classes('ach-workbench-body'):
                with ui.element('div').classes('ach-workbench-main') as graph_host:
                    self._graph_host_el = graph_host
                    self._render_palette(self._palette_dom_id)
                    with ui.element('div').classes(
                        'ach-sidebar-resizer'
                    ).props(
                        'role=separator aria-orientation=vertical '
                        'title="Drag to resize node sidebar. '
                        'Double-click to reset width."'
                    ) as resize_handle:
                        resize_handle.on(
                            'pointerdown',
                            lambda _e: None,
                            js_handler=(
                                '(e) => {'
                                'window.__oeAcherion?.beginSidebarResize('
                                'e.currentTarget, e.clientX);'
                                'e.preventDefault();'
                                'e.stopPropagation();'
                                '}'
                            ),
                        )
                        resize_handle.on(
                            'dblclick',
                            lambda _e: None,
                            js_handler=(
                                '(e) => {'
                                'window.__oeAcherion?.resetSidebarWidth('
                                'e.currentTarget.closest(".ach-workbench")'
                                ');'
                                'e.preventDefault();'
                                'e.stopPropagation();'
                                '}'
                            ),
                        )
                        ui.element('div').classes('ach-sidebar-resizer-grip')
                    with ui.element('div').classes('ach-shell').props(
                        f'tabindex=0 id={self._viewport_dom_id}'
                    ) as viewport:
                        self._attach_viewport_events(viewport)
                        self._canvas_el = ui.element('div').classes(
                            'ach-canvas'
                        ).props(f'tabindex=0 id={self._canvas_dom_id}')
                        self._attach_canvas_events(self._canvas_el)
                        self._ctx_container_el = ui.element('div')
                        self._overlay_host_el = ui.element('div')
                with ui.element('div').classes('ach-code-pane') as code_host:
                    self._code_host_el = code_host
                    if self._build_code_view is not None:
                        self._build_code_view()

        self.force_redraw()
        self._apply_mode_visibility()
        self.apply_preferences_state()

    def force_redraw(self: Any) -> None:
        """Refresh graph UI again after layout settles."""
        self._redraw_revision += 1
        redraw_revision = self._redraw_revision

        def _redraw_once() -> None:
            if redraw_revision != self._redraw_revision:
                return
            if self._canvas_el is None:
                return
            self.refresh()
            self._apply_mode_visibility()
            self._sync_client_viewport()

        _redraw_once()
        self._schedule_client_callback(
            0.0,
            _redraw_once,
            name='acherion redraw next tick',
        )
        self._schedule_client_callback(
            0.18,
            _redraw_once,
            name='acherion redraw settle 180ms',
        )
        self._schedule_client_callback(
            0.45,
            _redraw_once,
            name='acherion redraw settle 450ms',
        )
        self._schedule_client_callback(
            0.9,
            _redraw_once,
            name='acherion redraw settle 900ms',
        )

    def set_mode(self: Any, mode: str) -> None:
        """Switch between graph and code workbench modes."""
        if mode not in {'graph', 'code'}:
            return
        if self._editor_mode == mode:
            self._apply_mode_visibility()
            return
        self._editor_mode = mode
        self._apply_mode_visibility()
        if mode == 'graph':
            self._sync_client_viewport()
        if self._on_mode_change is not None:
            self._on_mode_change(mode)

    def refresh(self: Any) -> None:
        """Re-render node cards and connection SVG."""
        self._normalize_graph()
        valid_ids = {n.node_id for n in self._graph.nodes}
        self._selected_node_ids &= valid_ids
        pure_pending = (
            self._pending_source_node_id.split('@', 1)[0]
            if self._pending_source_node_id
            else None
        )
        if pure_pending not in valid_ids:
            self._pending_source_node_id = None
        valid_conn_ids = {
            str(spec['connection_id']) for spec in self._connection_specs()
        }
        if self._selected_connection_id not in valid_conn_ids:
            self._selected_connection_id = None
        if self._canvas_el is None:
            return
        self._canvas_el.clear()
        self._canvas_el.style(
            f'height:{self._canvas_height()}px;'
            f'width:max(100%,{self._canvas_width()}px);'
            f'--ach-origin-x:{self._canvas_origin_x()}px;'
            f'--ach-origin-y:{self._canvas_origin_y()}px;'
        )
        with self._canvas_el:
            self._render_links()
            if not self._graph.nodes:
                ui.label(
                    'No nodes yet. Add components from the left sidebar.'
                ).classes('ach-empty text-sm').style(
                    f'position:absolute; left:{self._world_to_canvas_x(32)}px;'
                    f' top:{self._world_to_canvas_y(32)}px; width:360px;'
                )
                self._sync_client_viewport()
                return
            index_map = {
                node.node_id: index
                for index, node in enumerate(self._graph.nodes)
            }
            self._render_group_frames()
            for node in self._graph.nodes:
                self._render_node(node, index_map[node.node_id])
            self._sync_client_viewport()
        if self._ctx_container_el is not None:
            self._ctx_container_el.clear()
            with self._ctx_container_el:
                self._render_context_menu()

    def _apply_mode_visibility(self: Any) -> None:
        graph_active = self._editor_mode == 'graph'
        if self._graph_host_el is not None:
            if graph_active:
                self._graph_host_el.classes(remove='ach-mode-hidden')
            else:
                self._graph_host_el.classes(add='ach-mode-hidden')
        if self._code_host_el is not None:
            if graph_active:
                self._code_host_el.classes(add='ach-mode-hidden')
            else:
                self._code_host_el.classes(remove='ach-mode-hidden')
        if self._mode_toggle_btn is not None:
            self._mode_toggle_btn.set_text('Code' if graph_active else 'Graph')
            self._mode_toggle_btn.props(
                'icon=code' if graph_active else 'icon=account_tree'
            )

    def _ensure_css(self: Any) -> None:
        if self._css_injected:
            return
        self._css_injected = True
        ui.add_css(_ACH_CSS)

    def _run_client_javascript(
        self: Any,
        code: str,
        *,
        timeout: float = 1.0,
    ) -> None:
        client = cast('Client | None', getattr(self, '_client', None))
        if client is None:
            try:
                client = ui.context.client
            except RuntimeError:
                return
            self._client = client
        client.run_javascript(code, timeout=timeout)

    def _schedule_client_callback(
        self: Any,
        delay: float,
        callback: Callable[[], None],
        *,
        name: str,
    ) -> None:
        client = cast('Client | None', getattr(self, '_client', None))
        if client is None:
            try:
                client = ui.context.client
            except RuntimeError:
                return
            self._client = client

        async def _run_later() -> None:
            if delay > 0:
                await asyncio.sleep(delay)
            if getattr(client, '_deleted', False):
                return
            client.safe_invoke(callback)

        background_tasks.create(_run_later(), name=name)

    def _ensure_client_js(self: Any) -> None:
        if self._client_js_injected:
            return
        self._client_js_injected = True
        self._run_client_javascript(_ACH_CLIENT_JS)

    def _sync_client_viewport(self: Any) -> None:
        vp = self._viewport_dom_id
        self._run_client_javascript(
            f'(function(){{'
            f'const sync=()=>{{'
            f'const vp=document.getElementById({vp!r});'
            f'if(!vp)return;'
            f'window.__oeAcherion?.observeViewport(vp);'
            f'window.__oeAcherion?.syncViewport(vp);'
            f'}};'
            f'sync();'
            f'setTimeout(sync,0);'
            f'setTimeout(sync,80);'
            f'setTimeout(sync,250);'
            f'setTimeout(sync,700);'
            f'}})()'
        )

    def _notify_change(self: Any) -> None:
        self._clear_preview_runtime_state()
        self.refresh()
        self._update_hint()
        if self._on_change is not None:
            self._on_change()

    def _render_toolbar_menu(
        self: Any,
        label: str,
        items: list[tuple[str, Callable[[], None]]],
    ) -> None:
        label_token = label.lower().replace(' ', '-')
        button_dom_id = f'{self._frame_dom_id}-{label_token}-button'
        menu_dom_id = f'{self._frame_dom_id}-{label_token}-menu'
        with ui.button(label).props(
            f'flat dense no-caps color=white id={button_dom_id}'
        ).classes(
            'ach-menu-button'
        ):
            with ui.menu().props(
                'auto-close anchor=bottom left self=top left '
                f'offset=[0,4] id={menu_dom_id}'
            ).classes('ach-menubar-menu') as menu:
                menu.on(
                    'show',
                    lambda _e, bid=button_dom_id, mid=menu_dom_id: (
                        self._show_toolbar_menu(bid, mid)
                    ),
                )
                menu.on(
                    'hide',
                    lambda _e, mid=menu_dom_id: self._hide_toolbar_menu(mid),
                )
                for item_label, handler in items:
                    ui.menu_item(item_label, on_click=handler)

    def _show_toolbar_menu(
        self: Any,
        button_dom_id: str,
        menu_dom_id: str,
    ) -> None:
        self._set_menu_open(True)
        self._run_client_javascript(
            f'(function(){{'
            f'const align=()=>{{'
            f'const button=document.getElementById({button_dom_id!r});'
            f'const menu=document.getElementById({menu_dom_id!r});'
            f'if(!button||!menu)return;'
            f'menu.removeAttribute("data-ach-aligned");'
            f'const buttonRect=button.getBoundingClientRect();'
            f'const menuRect=menu.getBoundingClientRect();'
            f'const maxLeft=Math.max('
            f'8, window.innerWidth - menuRect.width - 8'
            f');'
            f'const left=Math.min(Math.max(buttonRect.left, 8), maxLeft);'
            f'menu.style.left=`${{Math.round(left)}}px`;'
            f'menu.style.top=`${{Math.round(buttonRect.bottom + 4)}}px`;'
            f'menu.style.right="auto";'
            f'menu.style.transformOrigin="top left";'
            f'menu.setAttribute("data-ach-aligned", "true");'
            f'}};'
            f'requestAnimationFrame(align);'
            f'setTimeout(align,0);'
            f'setTimeout(align,80);'
            f'}})()'
        )

    def _hide_toolbar_menu(self: Any, menu_dom_id: str) -> None:
        self._set_menu_open(False)
        self._run_client_javascript(
            f'(function(){{'
            f'const menu=document.getElementById({menu_dom_id!r});'
            f'if(!menu)return;'
            f'menu.removeAttribute("data-ach-aligned");'
            f'}})()'
        )

    def _set_menu_open(self: Any, is_open: bool) -> None:
        frame_id = self._frame_dom_id
        class_name = 'ach-workbench-menu-open'
        action = 'add' if is_open else 'remove'
        self._run_client_javascript(
            f'(function(){{'
            f'const frame=document.getElementById({frame_id!r});'
            f'if(!frame)return;'
            f'frame.classList[{action!r}]({class_name!r});'
            f'}})()'
        )

    def _render_workbench_menu_bar(self: Any) -> None:
        with ui.element('div').classes('ach-menubar'):
            with ui.element('div').classes('ach-menubar-left'):
                with ui.row().classes('ach-menubar-menus'):
                    self._render_toolbar_menu(
                        'File',
                        [
                            ('Preferences', self._open_preferences_dialog),
                            ('Compile to Code', self._apply_to_code),
                            ('Refresh Canvas', self.refresh),
                        ],
                    )
                    self._render_toolbar_menu(
                        'Edit',
                        [
                            ('Clear Selection', self._clear_selection),
                            ('Delete Selection', self._delete_current_selection),
                        ],
                    )
                    self._render_toolbar_menu(
                        'View',
                        [
                            ('Zoom In', lambda: self._zoom_view(1.12)),
                            ('Zoom Out', lambda: self._zoom_view(0.89)),
                            ('Reset View', self._reset_view),
                        ],
                    )
                    self._render_toolbar_menu(
                        'Window',
                        [('Toggle Full Screen', self._toggle_full_screen)],
                    )
                    self._render_toolbar_menu(
                        'Tools',
                        [
                            ('Center on Graph', self._center_on_graph),
                            ('Sync Viewport', self._sync_client_viewport),
                        ],
                    )
                    self._render_toolbar_menu(
                        'Help',
                        [
                            ('Show Shortcuts', self._show_shortcuts_help),
                            ('Show Wiring Tips', self._show_wiring_help),
                        ],
                    )
                    ui.element('div').classes('ach-menubar-separator')
                    self._mode_toggle_btn = ui.button(
                        'Code',
                        icon='code',
                        on_click=lambda: self.set_mode(
                            'code' if self._editor_mode == 'graph' else 'graph'
                        ),
                    ).props('flat dense no-caps color=white').classes(
                        'ach-mode-button'
                    )
                    ui.button(
                        'Compile',
                        icon='play_arrow',
                        on_click=self._apply_to_code,
                    ).props('unelevated dense no-caps').classes(
                        'ach-compile-button'
                    )
                    ui.button(
                        'Validate',
                        icon='check_circle',
                        on_click=self._validate_current,
                    ).props('flat dense no-caps color=white').classes(
                        'ach-validate-button'
                    )
            with ui.element('div').classes('ach-menubar-right'):
                with ui.element('div').classes('ach-statusbar'):
                    ui.label('Status').classes('ach-statusbar-label')
                    self._hint_label = ui.label('').classes(
                        'ach-statusbar-text'
                    )
                    self._update_hint()

    def _run_viewport_command(self: Any, command: str) -> None:
        viewport_id = self._viewport_dom_id
        self._run_client_javascript(
            f'(function(){{'
            f'const vp=document.getElementById({viewport_id!r});'
            f'const h=window.__oeAcherion;'
            f'if(!vp||!h)return;'
            f'{command}'
            f'}})()'
        )

    def _clear_selection(self: Any) -> None:
        if not self._selected_node_ids and self._selected_connection_id is None:
            return
        self._selected_node_ids.clear()
        self._selected_connection_id = None
        self.refresh()
        self._update_hint()

    def _delete_current_selection(self: Any) -> None:
        if self._selected_node_ids:
            to_delete = set(self._selected_node_ids)
            self._selected_node_ids.clear()
            self._delete_nodes_batch(to_delete)
            return
        if self._selected_connection_id is not None:
            self._delete_selected_connection()
            return
        self._notify_ui(
            'Select nodes or a connection first.',
            type='warning',
        )

    def _zoom_view(self: Any, factor: float) -> None:
        self._run_viewport_command(f'h.zoomViewport(vp,{factor});')

    def _reset_view(self: Any) -> None:
        self._run_viewport_command('h.resetViewport(vp);')

    def _center_on_graph(self: Any) -> None:
        self._run_viewport_command('h.centerGraph(vp);')

    def _toggle_full_screen(self: Any) -> None:
        frame_id = self._frame_dom_id
        self._run_client_javascript(
            f'(function(){{'
            f'const frame=document.getElementById({frame_id!r});'
            f'const h=window.__oeAcherion;'
            f'if(!frame||!h)return;'
            f'h.toggleFullscreen(frame);'
            f'}})()'
        )

    def _show_shortcuts_help(self: Any) -> None:
        self._notify_ui(
            'Shortcuts: '
            f'{self._shortcut_display_binding("toggle_selection")} toggles '
            'node selection. '
            f'{self._shortcut_display_binding("box_select")} starts box '
            'selection. '
            f'{self._shortcut_display_binding("delete_selection_primary")} '
            'deletes selected nodes or connections. '
            f'{self._shortcut_display_binding("clear_selection")} clears '
            'selection.',
            type='info',
        )

    def _show_wiring_help(self: Any) -> None:
        self._notify_ui(
            'Click an output pin, then an input pin to connect. White exec '
            'wires control order. Coloured pins carry data values.',
            type='info',
        )

    def _notify_ui(
        self: Any,
        message: str,
        *,
        type: Literal[
            'positive',
            'negative',
            'warning',
            'info',
            'ongoing',
        ] = 'info',
    ) -> None:
        """Show a notification using stable host context when possible."""
        host = self._overlay_host_el or self._canvas_el or self._ctx_container_el
        if host is not None:
            with host:
                ui.notify(message, type=type)
            return
        ui.notify(message, type=type)

    def _persist_only(self: Any) -> None:
        """Persist data without re-rendering node cards."""
        if self._on_change is not None:
            self._on_change()

    def _validate_current(self: Any) -> None:
        if self._on_validate is None:
            self._notify_ui(
                'Validation is not available for this Logic Studio instance.',
                type='warning',
            )
            return
        self._on_validate()

    def set_status_message(
        self: Any,
        message: str,
        *,
        negative: bool = False,
    ) -> None:
        """Persist and display a workbench status message."""
        self._status_override_text = message
        self._status_override_negative = negative
        self._update_hint()

    def clear_status_message(self: Any) -> None:
        """Return the workbench status bar to contextual hints."""
        self._status_override_text = None
        self._status_override_negative = False
        self._update_hint()

    def _apply_to_code(self: Any) -> None:
        if self._on_apply_to_code is not None:
            self._on_apply_to_code()

    def _update_hint(self: Any, message: str | None = None) -> None:
        if self._hint_label is None:
            return
        negative = False
        if message is None:
            if self._status_override_text is not None:
                message = self._status_override_text
                negative = self._status_override_negative
            elif self._drag_node_id is not None:
                message = 'Dragging node. Release to place it.'
            elif self._selected_node_ids:
                count = len(self._selected_node_ids)
                message = (
                    f'{count} node{"s" if count > 1 else ""} selected. '
                    f'{self._shortcut_display_binding("toggle_selection")} '
                    'toggles selection. Drag any selected node to move '
                    'group. Right-click for group options. '
                    f'{self._shortcut_display_binding("delete_selection_primary")} '
                    'removes selection. '
                    f'{self._shortcut_display_binding("clear_selection")} '
                    'clears it.'
                )
            elif self._selected_connection_id is not None:
                message = (
                    'Connection selected. Press '
                    f'{self._shortcut_display_binding("delete_selection_primary")} '
                    'to remove it, or right-click the wire.'
                )
            elif self._pending_source_node_id is not None:
                message = (
                    f'Connecting from '
                    f'{self._source_label(self._pending_source_node_id)}. '
                    'Click an input pin.'
                )
            else:
                message = (
                    'Drag node headers to place boxes. Click an output pin, '
                    'then an input pin to connect components.'
                )
        self._hint_label.text = message
        self._hint_label.classes(remove='text-positive text-negative')
        if negative:
            self._hint_label.classes(add='text-negative')
        elif self._status_override_text is not None:
            self._hint_label.classes(add='text-positive')


"""Graph-operations mixin for AcherionDesigner.

Provides system-node sync, node CRUD, connection management, and all
methods that mutate the graph model. Intended to be inherited by
AcherionDesigner only.
"""
from typing import Any, cast

import acherion.node_behaviors as acherion_node_behaviors

from acherion.constants import (
    _GRID_SNAP_SIZE,
    _GROUP_COLOURS,
    _MANUAL_X,
    _MANUAL_Y,
    _NODE_STEP,
    _NODE_TOP,
    _SOURCE_X,
)
from acherion.events import (
    AcherionExternalEvent,
    default_acherion_external_events,
    normalize_acherion_external_events,
)
from acherion.registry import (
    get_acherion_node_definition,
    is_acherion_system_sink_kind,
    is_acherion_system_source_kind,
)
from acherion.model import (
    AcherionNode,
    _default_node,
    _system_node_id,
)


_AcherionLayoutMetric = dict[str, float | int]
_AcherionLayoutMetrics = dict[str, _AcherionLayoutMetric]
_AcherionLayoutBounds = dict[str, float]
_AcherionLayoutTargets = dict[str, tuple[int, int]]
_AcherionLayoutCommandResult = tuple[bool, str | _AcherionLayoutTargets]

_LAYOUT_COMMAND_SPECS: dict[str, tuple[str, object]] = {
    'align_top': ('align', ('y', 'top')),
    'align_middle': ('align', ('y', 'center')),
    'align_bottom': ('align', ('y', 'bottom')),
    'align_left': ('align', ('x', 'left')),
    'align_center': ('align', ('x', 'center')),
    'align_right': ('align', ('x', 'right')),
    'straighten_connections': ('straighten', None),
    'distribute_horizontally': ('distribute', 'x'),
    'distribute_vertically': ('distribute', 'y'),
    'stack_horizontally': ('stack', 'x'),
    'stack_vertically': ('stack', 'y'),
}

_LAYOUT_COMMAND_MESSAGES: dict[str, str] = {
    'align_top': 'Aligned selected nodes to the top edge.',
    'align_middle': 'Aligned selected nodes to the vertical middle.',
    'align_bottom': 'Aligned selected nodes to the bottom edge.',
    'align_left': 'Aligned selected nodes to the left edge.',
    'align_center': 'Aligned selected nodes to the horizontal center.',
    'align_right': 'Aligned selected nodes to the right edge.',
    'straighten_connections': 'Straightened the selected node layout.',
    'distribute_horizontally': 'Distributed selected nodes horizontally.',
    'distribute_vertically': 'Distributed selected nodes vertically.',
    'stack_horizontally': 'Stacked selected nodes horizontally.',
    'stack_vertically': 'Stacked selected nodes vertically.',
}


class _GraphOpsMixin:
    """Graph-mutation methods for AcherionDesigner."""

    _CENTER_ADD_OFFSET = 48
    _CENTER_ADD_MAX_ATTEMPTS = 64
    _STACK_GAP = _GRID_SNAP_SIZE

    _selected_connection_id: str | None = None

    @staticmethod
    def _snap_grid_value(value: int | float) -> int:
        return int(round(float(value) / _GRID_SNAP_SIZE) * _GRID_SNAP_SIZE)

    def _snap_grid_point(
        self: Any,
        x: int | float,
        y: int | float,
    ) -> tuple[int, int]:
        return (
            self._snap_grid_value(x),
            self._snap_grid_value(y),
        )

    # --- node predicates ---------------------------------------------------

    @staticmethod
    def _is_system_node(node: AcherionNode) -> bool:
        return bool(node.params.get('system_node'))

    def _is_system_source_node(self: Any, node: AcherionNode) -> bool:
        return self._is_system_node(node) and is_acherion_system_source_kind(
            node.kind
        )

    def _is_system_sink_node(self: Any, node: AcherionNode) -> bool:
        return self._is_system_node(node) and is_acherion_system_sink_kind(
            node.kind
        )

    @staticmethod
    def _pure_node_id(source_id: str) -> str:
        """Strip '@pin_index' suffix from a source id."""
        return source_id.split('@')[0] if source_id and '@' in source_id else (source_id or '')

    @staticmethod
    def _source_pin_index(source_id: str) -> int:
        """Return the output pin index encoded in a source id (default 0)."""
        if source_id and '@' in source_id:
            try:
                return int(source_id.split('@', 1)[1])
            except ValueError:
                return 0
        return 0

    def _manual_nodes(self: Any) -> list[AcherionNode]:
        return [
            n for n in self._graph.nodes if not self._is_system_node(n)
        ]

    def _external_events(self: Any) -> list[AcherionExternalEvent]:
        events: dict[str, AcherionExternalEvent]
        if self._host is not None:
            events = normalize_acherion_external_events(
                self._host.external_events()
            )
        else:
            events = default_acherion_external_events()
        for node in self._manual_nodes():
            definition = get_acherion_node_definition(node.kind)
            if definition is None:
                continue
            for event in definition.external_events(self, node):
                events[event.event_key] = event
        return list(normalize_acherion_external_events(events).values())

    def _sync_manual_schema_keys(self: Any) -> None:
        if self._host is not None:
            self._host.sync_manual_schema_keys(self)

    def _source_nodes(self: Any) -> list[AcherionNode]:
        return [
            n for n in self._graph.nodes if self._is_system_source_node(n)
        ]

    def _sink_nodes(self: Any) -> list[AcherionNode]:
        return [
            n for n in self._graph.nodes if self._is_system_sink_node(n)
        ]

    def _rebuild_graph(
        self: Any,
        manual_nodes: list[AcherionNode],
    ) -> None:
        self._graph.nodes = (
            self._source_nodes() + manual_nodes + self._sink_nodes()
        )
        self._ensure_function_box_entries()
        self._sync_function_box_ports()

    # --- position seeding --------------------------------------------------

    def _take_matching_node(  # type: ignore[return]
        self,
        *,
        consumed_ids: set[str],
        kind: str,
        key: str,
        param_key: str,
    ) -> AcherionNode | None:
        owner = cast(Any, self)
        expected_id = _system_node_id(kind, key)
        for node in owner._graph.nodes:
            if node.node_id in consumed_ids:
                continue
            if node.node_id == expected_id:
                consumed_ids.add(node.node_id)
                matched_node: AcherionNode = node
                return matched_node
            if node.kind != kind:
                continue
            if str(node.params.get(param_key) or '').strip() != key:
                continue
            consumed_ids.add(node.node_id)
            matched_node = node
            return matched_node
        return None

    def _seed_position(
        self: Any,
        node: AcherionNode,
        *,
        group: str,
        index: int,
    ) -> None:
        if bool(node.params.get('manual_position')):
            return
        if group == 'source':
            node.params.setdefault('dock', 'left')
            node.params.setdefault('y', _NODE_TOP + (index * _NODE_STEP))
            return
        if group == 'sink':
            node.params.setdefault('dock', 'right')
            node.params.setdefault('y', _NODE_TOP + (index * _NODE_STEP))
            return
        node.params.setdefault('dock', 'free')
        node.params.setdefault('x', _MANUAL_X)
        node.params.setdefault('y', _MANUAL_Y + (index * _NODE_STEP))

    def _rects_overlap(
        self: Any,
        *,
        left: int,
        top: int,
        width: int,
        height: int,
        other_left: int,
        other_top: int,
        other_width: int,
        other_height: int,
        padding: int = 24,
    ) -> bool:
        return not (
            (left + width + padding) <= other_left
            or left >= (other_left + other_width + padding)
            or (top + height + padding) <= other_top
            or top >= (other_top + other_height + padding)
        )

    def _visible_node_collision(
        self: Any,
        node: AcherionNode,
        *,
        left: int,
        top: int,
    ) -> bool:
        width = self._node_width(node)
        height = self._node_height(node)
        owner = cast(Any, self)
        for other_node in owner._graph.nodes:
            if other_node.node_id == node.node_id:
                continue
            if (
                owner._is_function_input(other_node)
                or owner._is_function_output(other_node)
                or owner._is_function_entry(other_node)
            ):
                continue
            if self._rects_overlap(
                left=left,
                top=top,
                width=width,
                height=height,
                other_left=self._node_world_left(other_node),
                other_top=self._node_world_top(other_node),
                other_width=self._node_width(other_node),
                other_height=self._node_height(other_node),
            ):
                return True
        return False

    def _resolve_centered_manual_position(
        self: Any,
        node: AcherionNode,
        *,
        center_x: int,
        center_y: int,
    ) -> tuple[int, int]:
        width = self._node_width(node)
        height = self._node_height(node)
        base_left = center_x - max(1, width // 2)
        base_top = center_y - max(1, height // 2)
        for attempt in range(self._CENTER_ADD_MAX_ATTEMPTS):
            offset = attempt * self._CENTER_ADD_OFFSET
            candidate_left = base_left + offset
            candidate_top = base_top + offset
            if not self._visible_node_collision(
                node,
                left=candidate_left,
                top=candidate_top,
            ):
                return candidate_left, candidate_top
        return base_left, base_top

    # --- system node sync --------------------------------------------------

    def _sync_system_nodes(self: Any) -> None:  # noqa: C901
        if self._host is not None:
            self._host.sync_system_nodes(self)
        return

    # --- node CRUD ---------------------------------------------------------

    def _add_node(
        self: Any,
        kind: str,
        center_x: int | None = None,
        center_y: int | None = None,
    ) -> None:
        manual_nodes = self._manual_nodes()
        new_node = _default_node(kind)
        if center_x is not None and center_y is not None:
            left, top = self._resolve_centered_manual_position(
                new_node,
                center_x=center_x,
                center_y=center_y,
            )
            new_node.params['x'] = left
            new_node.params['y'] = top
            new_node.params['dock'] = 'free'
            new_node.params['manual_position'] = True
            new_node.params['parent_function'] = self._containing_function_box_id(
                center_x,
                center_y,
            )
        else:
            self._seed_position(
                new_node, group='manual', index=len(manual_nodes)
            )
            target_box_id = self._selected_function_box_target()
            if target_box_id:
                self._place_node_inside_function_box(new_node, target_box_id)
        self._ensure_custom_function_entry(new_node)
        manual_nodes.append(new_node)
        self._rebuild_graph(manual_nodes)
        self._notify_change()

    def _add_node_at_position(
        self: Any, kind: str, x: int, y: int
    ) -> None:
        """Add a node of kind at a specific canvas position from a drag-drop."""
        manual_nodes = self._manual_nodes()
        new_node = _default_node(kind)
        x, y = self._snap_grid_point(x, y)
        new_node.params['x'] = x
        new_node.params['y'] = y
        new_node.params['dock'] = 'free'
        new_node.params['manual_position'] = True
        center_x = x + max(1, self._node_width(new_node) // 2)
        center_y = y + max(1, self._node_height(new_node) // 2)
        new_node.params['parent_function'] = self._containing_function_box_id(
            center_x,
            center_y,
        )
        self._ensure_custom_function_entry(new_node)
        manual_nodes.append(new_node)
        self._rebuild_graph(manual_nodes)
        self._notify_change()

    def _move_node(self: Any, node_id: str, direction: int) -> None:
        manual_nodes = self._manual_nodes()
        old_index = next(
            (
                i for i, n in enumerate(manual_nodes)
                if n.node_id == node_id
            ),
            -1,
        )
        if old_index < 0:
            return
        new_index = max(0, min(len(manual_nodes) - 1, old_index + direction))
        if old_index == new_index:
            return
        node = manual_nodes.pop(old_index)
        manual_nodes.insert(new_index, node)
        self._rebuild_graph(manual_nodes)
        self._notify_change()

    def _selected_layout_nodes(self: Any) -> list[AcherionNode]:
        selected_ids = set(self._selected_node_ids)
        if not selected_ids:
            return []
        selected_nodes = [
            node for node in self._graph.nodes
            if node.node_id in selected_ids and not self._is_function_entry(node)
        ]
        selected_box_ids = {
            node.node_id
            for node in selected_nodes
            if self._is_function_box(node)
        }
        return [
            node
            for node in selected_nodes
            if not any(
                box_id != node.node_id
                and self._node_has_function_ancestor(node.node_id, box_id)
                for box_id in selected_box_ids
            )
        ]

    def _layout_metrics(
        self: Any,
        node: AcherionNode,
    ) -> _AcherionLayoutMetric:
        left = int(self._node_world_left(node))
        top = int(self._node_world_top(node))
        width = int(self._node_width(node))
        height = int(self._node_height(node))
        return {
            'left': left,
            'top': top,
            'right': left + width,
            'bottom': top + height,
            'width': width,
            'height': height,
            'center_x': left + (width / 2.0),
            'center_y': top + (height / 2.0),
        }

    def _function_box_descendants(
        self: Any,
        box_id: str,
    ) -> list[AcherionNode]:
        return [
            node
            for node in self._manual_nodes()
            if node.node_id != box_id
            and self._node_has_function_ancestor(node.node_id, box_id)
        ]

    @staticmethod
    def _median(values: list[float]) -> float:
        ordered = sorted(values)
        mid = len(ordered) // 2
        if len(ordered) % 2:
            return ordered[mid]
        return (ordered[mid - 1] + ordered[mid]) / 2.0

    def _apply_layout_positions(
        self: Any,
        nodes: list[AcherionNode],
        targets: _AcherionLayoutTargets,
    ) -> bool:
        if not targets:
            return False
        current_positions = {
            node.node_id: (
                int(self._node_world_left(node)),
                int(self._node_world_top(node)),
            )
            for node in self._graph.nodes
        }
        resolved_targets = {
            node_id: self._snap_grid_point(left, top)
            for node_id, (left, top) in targets.items()
        }
        for node in nodes:
            if not self._is_function_box(node):
                continue
            current_left, current_top = current_positions.get(
                node.node_id,
                (0, 0),
            )
            target_left, target_top = resolved_targets.get(
                node.node_id,
                (current_left, current_top),
            )
            delta_x = target_left - current_left
            delta_y = target_top - current_top
            if delta_x == 0 and delta_y == 0:
                continue
            for child in self._function_box_descendants(node.node_id):
                child_left, child_top = current_positions.get(
                    child.node_id,
                    (
                        int(self._node_world_left(child)),
                        int(self._node_world_top(child)),
                    ),
                )
                resolved_targets[child.node_id] = self._snap_grid_point(
                    child_left + delta_x,
                    child_top + delta_y,
                )

        changed = False
        moved_nodes: list[tuple[AcherionNode, int, int]] = []
        for node_id, (left, top) in resolved_targets.items():
            node = self._node_by_id(node_id)
            if node is None:
                continue
            current_left, current_top = current_positions.get(
                node_id,
                (left, top),
            )
            if current_left == left and current_top == top:
                continue
            node.params['x'] = left
            node.params['y'] = top
            node.params['dock'] = 'free'
            node.params['manual_position'] = True
            moved_nodes.append((node, left, top))
            changed = True
        if not changed:
            return False
        for moved_node, left, top in moved_nodes:
            self._assign_node_to_containing_function_box(moved_node, left, top)
        self._notify_change()
        return True

    @staticmethod
    def _layout_selection_bounds(
        metrics: _AcherionLayoutMetrics,
    ) -> _AcherionLayoutBounds:
        left = min(float(data['left']) for data in metrics.values())
        top = min(float(data['top']) for data in metrics.values())
        right = max(float(data['right']) for data in metrics.values())
        bottom = max(float(data['bottom']) for data in metrics.values())
        return {
            'left': left,
            'top': top,
            'right': right,
            'bottom': bottom,
            'center_x': (left + right) / 2.0,
            'center_y': (top + bottom) / 2.0,
        }

    @staticmethod
    def _ordered_layout_nodes(
        nodes: list[AcherionNode],
        metrics: _AcherionLayoutMetrics,
        *,
        primary: str,
        secondary: str,
    ) -> list[AcherionNode]:
        return sorted(
            nodes,
            key=lambda node: (
                float(metrics[node.node_id][primary]),
                float(metrics[node.node_id][secondary]),
            ),
        )

    def _layout_axis_center_targets(
        self: Any,
        nodes: list[AcherionNode],
        metrics: _AcherionLayoutMetrics,
        *,
        axis: str,
        center_value: float,
    ) -> _AcherionLayoutTargets:
        targets: _AcherionLayoutTargets = {}
        for node in nodes:
            data = metrics[node.node_id]
            if axis == 'y':
                targets[node.node_id] = (
                    int(data['left']),
                    self._snap_grid_value(
                        center_value - (float(data['height']) / 2.0)
                    ),
                )
                continue
            targets[node.node_id] = (
                self._snap_grid_value(
                    center_value - (float(data['width']) / 2.0)
                ),
                int(data['top']),
            )
        return targets

    def _layout_align_targets(
        self: Any,
        nodes: list[AcherionNode],
        metrics: _AcherionLayoutMetrics,
        bounds: _AcherionLayoutBounds,
        *,
        axis: str,
        anchor: str,
    ) -> _AcherionLayoutTargets:
        if anchor == 'center':
            return self._layout_axis_center_targets(
                nodes,
                metrics,
                axis=axis,
                center_value=float(bounds['center_y' if axis == 'y' else 'center_x']),
            )

        targets: _AcherionLayoutTargets = {}
        for node in nodes:
            data = metrics[node.node_id]
            if axis == 'y':
                anchor_values = {
                    'top': float(bounds['top']),
                    'bottom': float(bounds['bottom']) - float(data['height']),
                }
                targets[node.node_id] = (
                    int(data['left']),
                    self._snap_grid_value(anchor_values[anchor]),
                )
                continue
            anchor_values = {
                'left': float(bounds['left']),
                'right': float(bounds['right']) - float(data['width']),
            }
            targets[node.node_id] = (
                self._snap_grid_value(anchor_values[anchor]),
                int(data['top']),
            )
        return targets

    def _layout_straighten_targets(
        self: Any,
        nodes: list[AcherionNode],
        metrics: _AcherionLayoutMetrics,
        bounds: _AcherionLayoutBounds,
    ) -> _AcherionLayoutTargets:
        horizontal = (
            (float(bounds['right']) - float(bounds['left']))
            >= (float(bounds['bottom']) - float(bounds['top']))
        )
        if horizontal:
            target_center = self._median([
                float(data['center_y']) for data in metrics.values()
            ])
            return self._layout_axis_center_targets(
                nodes,
                metrics,
                axis='y',
                center_value=target_center,
            )
        target_center = self._median([
            float(data['center_x']) for data in metrics.values()
        ])
        return self._layout_axis_center_targets(
            nodes,
            metrics,
            axis='x',
            center_value=target_center,
        )

    def _layout_distribute_targets(
        self: Any,
        nodes: list[AcherionNode],
        metrics: _AcherionLayoutMetrics,
        bounds: _AcherionLayoutBounds,
        *,
        axis: str,
    ) -> _AcherionLayoutCommandResult:
        if len(nodes) < 3:
            if axis == 'x':
                return (
                    False,
                    'Select at least three nodes to distribute horizontally.',
                )
            return (
                False,
                'Select at least three nodes to distribute vertically.',
            )

        if axis == 'x':
            ordered = self._ordered_layout_nodes(
                nodes,
                metrics,
                primary='left',
                secondary='top',
            )
            total_size = sum(
                float(metrics[node.node_id]['width']) for node in ordered
            )
            gap = max(
                0.0,
                (float(bounds['right']) - float(bounds['left']) - total_size)
                / (len(ordered) - 1),
            )
            cursor = float(bounds['left'])
            targets: _AcherionLayoutTargets = {}
            for node in ordered:
                data = metrics[node.node_id]
                targets[node.node_id] = (
                    self._snap_grid_value(cursor),
                    int(data['top']),
                )
                cursor += float(data['width']) + gap
            return (True, targets)

        ordered = self._ordered_layout_nodes(
            nodes,
            metrics,
            primary='top',
            secondary='left',
        )
        total_size = sum(
            float(metrics[node.node_id]['height']) for node in ordered
        )
        gap = max(
            0.0,
            (float(bounds['bottom']) - float(bounds['top']) - total_size)
            / (len(ordered) - 1),
        )
        cursor = float(bounds['top'])
        targets = {}
        for node in ordered:
            data = metrics[node.node_id]
            targets[node.node_id] = (
                int(data['left']),
                self._snap_grid_value(cursor),
            )
            cursor += float(data['height']) + gap
        return (True, targets)

    def _layout_stack_targets(
        self: Any,
        nodes: list[AcherionNode],
        metrics: _AcherionLayoutMetrics,
        bounds: _AcherionLayoutBounds,
        *,
        axis: str,
    ) -> _AcherionLayoutTargets:
        if axis == 'x':
            ordered = self._ordered_layout_nodes(
                nodes,
                metrics,
                primary='left',
                secondary='top',
            )
            cursor = float(bounds['left'])
            targets: _AcherionLayoutTargets = {}
            for node in ordered:
                data = metrics[node.node_id]
                targets[node.node_id] = self._snap_grid_point(
                    cursor,
                    float(bounds['center_y']) - (float(data['height']) / 2.0),
                )
                cursor += float(data['width']) + self._STACK_GAP
            return targets

        ordered = self._ordered_layout_nodes(
            nodes,
            metrics,
            primary='top',
            secondary='left',
        )
        cursor = float(bounds['top'])
        targets = {}
        for node in ordered:
            data = metrics[node.node_id]
            targets[node.node_id] = self._snap_grid_point(
                float(bounds['center_x']) - (float(data['width']) / 2.0),
                cursor,
            )
            cursor += float(data['height']) + self._STACK_GAP
        return targets

    def _layout_selected_nodes(
        self: Any,
        command: str,
    ) -> tuple[bool, str]:
        nodes = self._selected_layout_nodes()
        if len(nodes) < 2:
            return (False, 'Select at least two nodes first.')

        metrics: _AcherionLayoutMetrics = {
            node.node_id: self._layout_metrics(node)
            for node in nodes
        }
        bounds = self._layout_selection_bounds(metrics)
        command_spec = _LAYOUT_COMMAND_SPECS.get(command)
        if command_spec is None:
            return (False, 'Unknown layout command.')

        mode, config = command_spec
        dispatchers = {
            'align': lambda raw: (
                True,
                self._layout_align_targets(
                    nodes,
                    metrics,
                    bounds,
                    axis=cast(tuple[str, str], raw)[0],
                    anchor=cast(tuple[str, str], raw)[1],
                ),
            ),
            'straighten': lambda _raw: (
                True,
                self._layout_straighten_targets(nodes, metrics, bounds),
            ),
            'distribute': lambda raw: self._layout_distribute_targets(
                nodes,
                metrics,
                bounds,
                axis=cast(str, raw),
            ),
            'stack': lambda raw: (
                True,
                self._layout_stack_targets(
                    nodes,
                    metrics,
                    bounds,
                    axis=cast(str, raw),
                ),
            ),
        }
        ok, payload = dispatchers[mode](config)
        if not ok:
            return (False, cast(str, payload))
        targets = cast(_AcherionLayoutTargets, payload)

        if not self._apply_layout_positions(nodes, targets):
            return (True, 'Selection already matches that layout.')
        return (
            True,
            _LAYOUT_COMMAND_MESSAGES.get(
                command,
                'Updated selected node layout.',
            ),
        )

    def _delete_node(self: Any, node_id: str) -> None:
        node = self._node_by_id(node_id)
        if node is not None and self._is_function_entry(node):
            return
        manual_nodes = self._manual_nodes()
        if any(node.node_id == node_id and self._is_function_box(node) for node in manual_nodes):
            child_ids = {
                node.node_id for node in manual_nodes
                if self._function_parent_id(node) == node_id
            }
            child_ids.add(node_id)
            self._delete_nodes_batch(child_ids)
            return
        manual_nodes = [
            n for n in self._manual_nodes() if n.node_id != node_id
        ]
        self._rebuild_graph(manual_nodes)
        self._cleanup_custom_function_entries()
        for node in manual_nodes:
            if self._is_function_box(node):
                self._prune_box_io_metadata(node)
        self._cleanup_empty_groups()
        self._notify_change()

    def _delete_nodes_batch(self: Any, node_ids: set[str]) -> None:
        if not node_ids:
            return
        manual_nodes = self._manual_nodes()
        expanded_ids = {
            node_id for node_id in node_ids
            if not any(
                node.node_id == node_id and self._is_function_entry(node)
                for node in manual_nodes
            )
        }
        for node in manual_nodes:
            if node.node_id in node_ids and self._is_function_box(node):
                expanded_ids.update(
                    child.node_id
                    for child in manual_nodes
                    if self._function_parent_id(child) == node.node_id
                )
        manual_nodes = [
            n for n in manual_nodes
            if n.node_id not in expanded_ids
        ]
        self._rebuild_graph(manual_nodes)
        self._cleanup_custom_function_entries()
        for node in manual_nodes:
            if self._is_function_box(node):
                self._prune_box_io_metadata(node)
        self._cleanup_empty_groups()
        self._notify_change()

    # --- group management -------------------------------------------------

    def _next_group_colour(self: Any) -> str:
        """Return next unused colour from the palette."""
        used = set(self._graph.groups.values())
        for colour in _GROUP_COLOURS:
            if colour not in used:
                return str(colour)
        return str(_GROUP_COLOURS[
            len(self._graph.groups) % len(_GROUP_COLOURS)
        ])

    def _create_group(self: Any, name: str, node_ids: set[str]) -> None:
        """Create a named group and assign selected nodes to it."""
        name = name.strip()
        if not name or name in self._graph.groups:
            return
        self._graph.groups[name] = self._next_group_colour()
        for node in self._graph.nodes:
            if node.node_id in node_ids:
                node.params['group'] = name
        self._notify_change()

    def _add_nodes_to_group(self: Any, name: str, node_ids: set[str]) -> None:
        """Move nodes into an existing group."""
        if name not in self._graph.groups:
            return
        for node in self._graph.nodes:
            if node.node_id in node_ids:
                node.params['group'] = name
        self._notify_change()

    def _remove_nodes_from_group(self: Any, node_ids: set[str]) -> None:
        """Remove nodes from their current group; prune empty groups."""
        for node in self._graph.nodes:
            if node.node_id in node_ids:
                node.params.pop('group', None)
        self._cleanup_empty_groups()
        self._notify_change()

    def _cleanup_empty_groups(self: Any) -> None:
        """Delete groups that have no member nodes."""
        members: set[str] = {
            str(n.params['group'])
            for n in self._graph.nodes
            if n.params.get('group')
        }
        self._graph.groups = {
            k: v
            for k, v in self._graph.groups.items()
            if k in members
        }

    def _extract_function_from_selection(
        self: Any,
        name: str,
        *,
        return_node_id: str | None = None,
    ) -> tuple[bool, str]:
        """Extract selected nodes into a composite function box."""
        selected_ids = set(self._selected_node_ids)
        if not selected_ids:
            return (False, 'Select nodes first.')

        selected_nodes = [
            node for node in self._manual_nodes()
            if node.node_id in selected_ids
        ]
        if len(selected_nodes) != len(selected_ids):
            return (False, 'Only manual nodes can be extracted into a function.')

        if any(self._is_function_box(node) for node in selected_nodes):
            return (False, 'Selecting a function box for extraction is not supported.')
        if any(
            self._is_function_input(node) or self._is_function_output(node)
            for node in selected_nodes
        ):
            return (False, 'Function input/output nodes cannot be re-extracted.')
        if any(self._is_function_entry(node) for node in selected_nodes):
            return (False, 'Function entry nodes cannot be re-extracted.')
        parent_ids = {
            self._function_parent_id(node) for node in selected_nodes
        }
        if len(parent_ids) > 1:
            return (
                False,
                'Select nodes from the same function scope.',
            )
        common_parent_id = next(iter(parent_ids), '')

        unsupported_kinds = {
            'for_each',
            'collect',
            'branch_route',
        }
        bad_kinds = sorted({
            node.kind
            for node in selected_nodes
            if (
                node.kind in unsupported_kinds
                or acherion_node_behaviors.function_box_unsupported(node)
            )
        })
        if bad_kinds:
            return (
                False,
                'Selection contains unsupported node kinds: '
                + ', '.join(bad_kinds),
            )

        function_slug = self._sanitize_identifier(name, 'function_box')
        manual_nodes = self._manual_nodes()
        insert_index = min(
            index
            for index, node in enumerate(manual_nodes)
            if node.node_id in selected_ids
        )
        box = _default_node('function_box')
        box.title = name.strip() or function_slug.replace('_', ' ').title()
        box.params['function_name'] = function_slug
        box.params['parent_function'] = common_parent_id
        box.params['manual_position'] = True
        box.params['dock'] = 'free'
        box.params['x'] = min(
            self._node_world_left(node) for node in selected_nodes
        ) - 48
        box.params['y'] = min(
            self._node_world_top(node) for node in selected_nodes
        ) - 72

        for node in selected_nodes:
            node.params['parent_function'] = box.node_id
            node.params.pop('group', None)

        new_manual_nodes = manual_nodes[:]
        new_manual_nodes.insert(insert_index, box)

        self._rebuild_graph(new_manual_nodes)
        self._selected_node_ids = {box.node_id}
        self._notify_change()
        return (
            True,
            f'Created function box {function_slug}. Boundary wires now define its inputs and outputs.',
        )



